/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is DogStar SOFTWARE PiXML Public Module.
 *
 * The Initial Developer of the Original Code is
 * James Tyson, DogStar SOFTWARE <james@thedogstar.org>.
 * Portions created by the Initial Developer are Copyright (C) 2005
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   None.
 *
 * ***** END LICENSE BLOCK ***** */

//! Parse some PiJAX XML back into a Pike data structure.
//!
//! @param xml
//!   An XML string containing the PiJAX XML.
mixed parse(string xml) {
  object root = Parser.XML.NSTree.parse_input(xml);
  object root_element;
  foreach(root->get_children(), object child)
    if ((child->get_node_type() == Parser.XML.Tree.XML_ELEMENT) &&
	(child->get_ns() == .Serialise->pike_xml_namespace())) {
      root_element = child;
      break;
    }
  return low_parse(root_element);
}

static mixed low_parse(object element) {
  mixed val;
  switch(element->get_tag_name()) {
    case "float":
      val = low_parse_float(element);
      break;
    case "integer":
      val = low_parse_int(element);
      break;
    case "string":
      val = low_parse_string(element);
      break;
    case "multiset":
      val = low_parse_multiset(element);
      break;
    case "array":
      val = low_parse_array(element);
      break;
    case "mapping":
      val = low_parse_mapping(element);
      break;
#if constant(Image.PNG.decode)
    case "image":
      val = low_parse_image(element);
      break;
#endif
    case "exception":
    default:
      throw(Error.Generic(sprintf("Unknown element of type %O\n", element->get_tag_name())));
      break;
  }
  return val;
}

static float low_parse_float(object element) {
  if (element->get_ns_attributes() &&
      element->get_ns_attributes()[.Serialise->pike_xml_namespace()] &&
      element->get_ns_attributes()[.Serialise->pike_xml_namespace()]["value"]) {
    return (float)element->get_ns_attributes()[.Serialise->pike_xml_namespace()]["value"];
  }
  else
    throw(Error.Generic("Error in <pike:float /> tag, no \"value\" attribute.\n"));
}

static int low_parse_int(object element) {
  if (element->get_ns_attributes() &&
      element->get_ns_attributes()[.Serialise->pike_xml_namespace()] &&
      element->get_ns_attributes()[.Serialise->pike_xml_namespace()]["value"]) {
    return (int)element->get_ns_attributes()[.Serialise->pike_xml_namespace()]["value"];
  }
  else
    throw(Error.Generic("Error in <pike:int /> tag, no \"value\" attribute.\n"));
}

static string low_parse_string(object element) {
  if (element->get_ns_attributes() &&
      element->get_ns_attributes()[.Serialise->pike_xml_namespace()] &&
      element->get_ns_attributes()[.Serialise->pike_xml_namespace()]["encoding"]) {
    array text = ({});
    foreach(element->get_children(), object child)
      if (child->get_node_type() == Parser.XML.Tree.XML_TEXT)
	text += ({ child });
    switch(element->get_ns_attributes()[.Serialise->pike_xml_namespace()]["encoding"]) {
      case "uue":
	return MIME.decode_uue(text->get_text() * "");
      case "base64":
	return MIME.decode_base64(text->get_text() * "");
      case "cdata":
	return text->get_text() * "";
      default:
	throw(Error.Generic(sprintf("Unknown script encoding %O\n", element->get_ns_attributes()[.Serialise->pike_xml_namespace()]["encoding"])));
    }
  }
  else
    throw(Error.Generic("Unable to parse string, encoding attribute not present\n"));
}

static multiset low_parse_multiset(object element) {
  if (element->get_ns_attributes() &&
      element->get_ns_attributes()[.Serialise->pike_xml_namespace()] &&
      element->get_ns_attributes()[.Serialise->pike_xml_namespace()]["elements"]) {
    int elements = (int)element->get_ns_attributes()[.Serialise->pike_xml_namespace()]["elements"];
    if (elements == 0)
      return (<>);
    else {
      multiset val = (<>);
      foreach(element->get_children(), object child)
	if ((child->get_ns() == .Serialise->pike_xml_namespace()) &&
	    (child->get_tag_name() == "element")) {
	  // This is a pike:element tag.
	  foreach(child->get_children(), object _child) 
	    if (_child->get_ns() == .Serialise->pike_xml_namespace())
	      val += (< low_parse(_child) >);
	}
      if (sizeof(val) == elements)
	return val;
      else
	throw(Error.Generic(sprintf("Parse of multiset revealed %d elements, should have been %d\n", sizeof(val), elements)));
    }
  }
  else
    throw(Error.Generic("Unable to parse multiset tag, elements attribute not present\n"));
}

static array low_parse_array(object element) {
  if (element->get_ns_attributes() &&
      element->get_ns_attributes()[.Serialise->pike_xml_namespace()] &&
      element->get_ns_attributes()[.Serialise->pike_xml_namespace()]["elements"]) {
    int elements = (int)element->get_ns_attributes()[.Serialise->pike_xml_namespace()]["elements"];
    if (elements == 0) 
      return ({});
    else {
      array val = ({});
      foreach(element->get_children(), object child)
	if ((child->get_ns() == .Serialise->pike_xml_namespace()) &&
	    (child->get_tag_name() == "element")) {
	  // This is a pike:element tag.
	  foreach(child->get_children(), object _child) 
	    if (_child->get_ns() == .Serialise->pike_xml_namespace())
	      val += ({ low_parse(_child) });
	}
      if (sizeof(val) == elements)
	return val;
      else
	throw(Error.Generic(sprintf("Parse of array revealed %d elements, should have been %d\n", sizeof(val), elements)));
    }
  }
  else
    throw(Error.Generic("Unable to parse array tag, elements attribute not present\n"));
}

static mapping low_parse_mapping(object element) {
  if (element->get_ns_attributes() &&
      element->get_ns_attributes()[.Serialise->pike_xml_namespace()] &&
      element->get_ns_attributes()[.Serialise->pike_xml_namespace()]["elements"]) {
    int elements = (int)element->get_ns_attributes()[.Serialise->pike_xml_namespace()]["elements"];
    if (elements == 0)
      return ([]);
    else {
      mapping val = ([]);
      foreach(element->get_children(), object child) {
	if ((child->get_ns() == .Serialise->pike_xml_namespace()) &&
	    (child->get_tag_name() == "map")) {
	  // This is a pike:map tag.
	  mixed _idx, _val;
	  foreach(child->get_children(), object _child) {
	    if ((_child->get_ns() == .Serialise->pike_xml_namespace()) &&
		(_child->get_tag_name() == "index" )) {
	      // This is a pike:index tag.
	      foreach(_child->get_children(), object __child) {
		if (__child->get_ns() == .Serialise->pike_xml_namespace()) {
		  _idx = low_parse(__child);
		}
	      }
	    }
	  }
	  foreach(child->get_children(), object _child) {
	    if ((_child->get_ns() == .Serialise->pike_xml_namespace()) &&
		(_child->get_tag_name() == "element" )) {
	      // This is a pike:element tag.
	      foreach(_child->get_children(), object __child) {
		if (__child->get_ns() == .Serialise->pike_xml_namespace()) {
		  _val = low_parse(__child);
		}
	      }
	    }
	  }
	  val[_idx] = _val;
	}
      }
      if (sizeof(val) == elements)
	return val;
      else
	throw(Error.Generic(sprintf("Parse of mapping revealed %d elements, should have been %d\n", sizeof(val), elements)));
    }
  }
  else
    throw(Error.Generic("Unable to parse mapping tag, elements attribute not present\n"));
}

#if constant(Image.PNG.decode)
static object low_parse_image(object element) {
  if (element->get_ns_attributes() &&
      element->get_ns_attributes()[.Serialise->pike_xml_namespace()] &&
      element->get_ns_attributes()[.Serialise->pike_xml_namespace()]["encoding"]) {
    array text = ({});
    string png;
    foreach(element->get_children(), object child)
      if (child->get_node_type() == Parser.XML.Tree.XML_TEXT)
	text += ({ child });
    switch(element->get_ns_attributes()[.Serialise->pike_xml_namespace()]["encoding"]) {
      case "uue":
	png = MIME.decode_uue(text->get_text() * "");
	break;
      case "base64":
	png = MIME.decode_base64(text->get_text() * "");
	break;
      case "cdata":
	png = text->get_text() * "";
	write(png);
	break;
      default:
	throw(Error.Generic(sprintf("Unknown script encoding %O\n", element->get_ns_attributes()[.Serialise->pike_xml_namespace()]["encoding"])));
	break;
    }
    if (stringp(png) && sizeof(png)) {
      object img;
      mixed err = catch(img = Image.PNG.decode(png));
      if (err)
	throw(err);
      else
	return img;
    }
    else 
      throw(Error.Generic("Empty PNG image data after decode\n"));
  }
  else 
    throw(Error.Generic("Unable to parse image, encoding attribute not present\n"));
}
#endif
