/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is DogStar SOFTWARE PiXML Public Module.
 *
 * The Initial Developer of the Original Code is
 * James Tyson, DogStar SOFTWARE <james@thedogstar.org>.
 * Portions created by the Initial Developer are Copyright (C) 2005
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   None.
 *
 * ***** END LICENSE BLOCK ***** */

static float  _xml_version  = 1.0;
static string _xml_encoding = "ISO-8859-1";
static string _pike_xml_ns  = "http://pike.ida.liu.se/";
static string _pike_xml_xmlns  = "pike";
static function _string_encoder = encode_cdata;
static string _string_encoder_name = "cdata";

//! Change the XML version number in the xml PI. Defaults to 1.0.
//! 
//! @param var
//!   Version number. Really should only be 1.0.
float xml_version(void|float ver) {
  if (ver && floatp(ver) && ver != _xml_version)
    _xml_version = ver;
  return _xml_version;
}

//! Change the XML encoding in the xml PI.
//! Defaults to "ISO-8859-1".
//! 
//! @param encoding
//!   The encoding.
string xml_encoding(void|string encoding) {
  if (encoding && stringp(encoding) && encoding != _xml_encoding)
    _xml_encoding = encoding;
  return _xml_encoding;
}

//! Change the string encoding using in output.
//! Defaults to using CDATA elements.
//! 
//! @param encoder
//!   Either "base64", "uue" or "cdata"
string string_encoding(void|string encoder) {
  if (encoder && stringp(encoder)) {
    switch(encoder) {
      case "uue":
	_string_encoder = MIME.encode_uue;
        _string_encoder_name = "uue";
	break;
      case "cdata":
	_string_encoder = encode_cdata;
	_string_encoder_name = "cdata";
	break;
      case "base64":
      default:
        _string_encoder = MIME.encode_base64;
	_string_encoder_name = "base64";
	break;
    }
  }
  return _string_encoder_name;
}

//! Change the default Pike namespace - don't do this unless you know
//! what you're doing.
//! Defaults to "http://pike.ida.liu.se/" and "pike".
//!
//! @param ns
//!   The URI of the new namespace.
//!
//! @param xmlns
//!   The XML namespace prefix.
string pike_xml_namespace(void|string ns, void|string xmlns) {
  if (ns && stringp(ns) && sizeof(ns))
    _pike_xml_ns = ns;
  if (xmlns && stringp(xmlns) && sizeof(xmlns))
    _pike_xml_xmlns = xmlns;
  return _pike_xml_ns;
}

//! Serialise the argument into XML.
//! 
//! @param data
//!   The data to be encoded.
//!
//! @seealso PiJAX.supported()
string serialise(mixed data) {
  string xml_pi = sprintf("<?xml version=\"%:01f\" encoding=\"%s\"?>\n", _xml_version, _xml_encoding);
  return xml_pi + low_serialise(data);
}

static string low_serialise(mixed data, void|int depth) {
  if (floatp(data))
    return floatx(data, depth);
  else if (intp(data))
    return intx(data, depth);
  else if (stringp(data))
    return stringx(data, depth);
  else if (multisetp(data))
    return multisetx(data, depth);
  else if (arrayp(data))
    return arrayx(data, depth);
  else if (mappingp(data))
    return mappingx(data, depth);
#if constant(Image.PNG.encode)
  else if (sprintf("%t", data) == "Image.Image")
    return imagex(data, depth);
#endif
  else
    return make_container("exception", ([]), sprintf("Error.  Don't know how to serialise %O\n", data), depth + 1);
}

static string floatx(float data, void|int depth) {
  return make_tag("float", ([ "value" : sprintf("%f", data) ]), depth);
}

static string intx(int data, void|int depth) {
  return make_tag("integer", ([ "value" : sprintf("%d", data) ]), depth);
}

static string stringx(int data, void|int depth) {
  return make_container("string", ([ "encoding" : _string_encoder_name ]), _string_encoder(data), depth);
}

static string multisetx(multiset data, void|int depth) {
  mapping args = ([]);
  string contents = "";
  foreach(indices(data), mixed idx)
    contents += make_container("element", ([]), low_serialise(idx, depth + 2), depth + 1);
  args += ([ "elements" : sizeof(data) ]);
  return make_container("multiset", args, contents, depth);
}

static string arrayx(array data, void|int depth) {
  mapping args = ([]);
  string contents = "";
  for(int i = 0; i < sizeof(data); i++)
    contents += make_container("element", ([ "index" : i ]), low_serialise(data[i], depth + 2), depth + 1);
  args += ([ "elements" : sizeof(data) ]);
  return make_container("array", args, contents, depth);
}

static string mappingx(mapping data, void|int depth) {
  mapping args = ([]);
  string contents = "";
  foreach(indices(data), mixed idx) {
    string index = make_container("index", ([]), low_serialise(idx, depth + 3), depth + 2);
    string element = make_container("element", ([]), low_serialise(data[idx], depth + 3), depth + 2);
    contents += make_container("map", ([]), index + element, depth + 1);
  }
  args += ([ "elements" : sizeof(data) ]);
  return make_container("mapping", args, contents, depth);
}

#if constant(Image.PNG.encode)
static string imagex(object data, void|int depth) {
  mapping args = ([]);
  args->encoding = "base64";
  args->mimetype = "image/png";
  return make_container("image", args, MIME.encode_base64(Image.PNG.encode(data)), depth);
}
#endif

static string make_tag(string tag, mapping args, void|int depth) {
  tag = lower_case(encode(tag));
  args = (mapping(string:string))args;
  foreach(indices(args), string idx) {
    string val = args[idx];
    m_delete(args, idx);
    args[encode(idx)] = encode(val);
  }
  if (!depth)
    args[sprintf("xmlns:%s", _pike_xml_xmlns)] = pike_xml_namespace();
  return sprintf(
      "%s<%s:%s%{ %s=\"%s\"%}/>\n",
      "  " * depth,
      _pike_xml_xmlns,
      tag,
      (array)args
    );
}

static string make_container(string tag, mapping args, string content, void|int depth) {
  tag = lower_case(encode(tag));
  args = (mapping(string:string))args;
  foreach(indices(args), string idx) {
    string val = args[idx];
    m_delete(args, idx);
    args[encode(idx)] = encode(val);
  }
  if (!depth)
    args[sprintf("xmlns:%s", _pike_xml_xmlns)] = pike_xml_namespace();
  if ((sizeof(content) > 40) || (sizeof(content) >= 2 && content[0..1] == "  "))
    return sprintf(
	"%s<%s:%s%{ %s=\"%s\"%}>\n%s%s</%s:%s>\n",
	"  " * depth, 
	_pike_xml_xmlns,
	tag,
	(array)args,
	content,
	"  " * depth,
	_pike_xml_xmlns,
	tag
      );
  else
    return sprintf(
	"%s<%s:%s%{ %s=\"%s\"%}>%s</%s:%s>\n",
	"  " * depth, 
	_pike_xml_xmlns,
	tag,
	(array)args,
	content,
	_pike_xml_xmlns,
	tag
      );
}

static string|array encode(function|array|string|int|float args) {
  array r = ({});
  array in;
  if (arrayp(args))
    in = args;
  else if (stringp(args))
    in = ({ args });
  else if (intp(args))
    in = ({ (string)args });
  else if (floatp(args))
    in = ({ sprintf("%:02f", args) });
  else if (functionp(args))
    in = ({ sprintf("%O", args) });
  foreach(in, mixed e) {
    if (intp(e) || floatp(e))
      e = (string)e;
    else if (!stringp(e))
      continue;

    r += ({ replace(
	e,
	({ "&", "<", ">", "\"", "\'", "\000" }),
	({ "&amp;", "&lt;", "&gt;", "&#34;", "&#39;", "&#0;" })
      ) });
  }
  if (arrayp(args))
    return r;
  else if (stringp(args))
    return r[0];
  else if (intp(args) || floatp(args))
    return r[0];
}

static string encode_cdata(string data) {
  return sprintf("<![CDATA[%s]]>", data);
}
